/*
 *  plugin-config.c - A plugin configuration tool
 *
 *  by Martin Stransky <stransky@redhat.com>
 *
 *  This utility is based on:
 *
 *  npw-config.c - nspluginwrapper configuration tool 
 *  nspluginwrapper (C) 2005-2007 Gwenole Beauchesne
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * This utility will install plugins for mozilla based browsers or
 * wrap them with the nspluginwrapper utility (if availabile)
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <stdarg.h>
#include <limits.h>
#include <errno.h>
#include <fnmatch.h>

#include <dlfcn.h>
#include <unistd.h>
#include <fcntl.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/fsuid.h>
#include <pwd.h>
#include <dirent.h>

#include <asm/types.h>

#include "plugin-path.h"
#include "plugin-dir.h"
#include "plugin-config.h"
#include "plugin-detection.h"

typedef enum {
  
  A_NONE,
  A_INSTALL,
  A_REMOVE,
  A_LIST,
  A_CHECK,
  A_HELP
   
} ACTIVE_ACTION;

typedef enum {
  
  PLUGIN_DIR,
  PLUGIN_DIR_WRAPPED
   
} SCANNED_DIR;

typedef struct _CONFIG {
  
  int verbose;
  int quiet;
  
  int global;
  int local;

  int config_32b;
  int config_64b;

  int force;

  int wrapper_32b;
  int wrapper_64b;

  ACTIVE_ACTION action;
  
} CONFIG;

CONFIG cfg;

char * exclude_default[] = 
{
  ".",
  "..",
  NPW_WRAPPER_PLUGIN
};

int is_excluded_default(char *p_name)
{
  int i;
  for(i = 0; i < sizeof(exclude_default) / sizeof(exclude_default[0]); i++) {
    if(!fnmatch(exclude_default[i],p_name,0))
      return(TRUE);    
  }
  return(FALSE);
}

inline char * tail_get(char * p_str, char delim)
{   
  char *p_last = NULL;
  
  while((p_str = strchr(p_str, delim))) {
    p_str++;
    p_last = p_str;    
  }  
  
  if(!p_last)
    p_last = p_str;
  
  return(p_last);
}

typedef enum {

  EXCLUDE_WRAP = 0,
  EXCLUDE_LINK = 1
  
} EXCLUDE_TYPE;

char * exclude_names[] = 
{
  "EXCLUDE_WRAP",
  "EXCLUDE_LINK"
};


static char *p_exclude_wrap = NULL;
static char *p_exclude_link = NULL;

void create_exclude_lists(void)
{
  p_exclude_wrap = getenv("IGNORE_WRAP");
  if(p_exclude_wrap)
    p_exclude_wrap = strdup(p_exclude_wrap);
  
  p_exclude_link = getenv("IGNORE_LINK");
  if(p_exclude_link)
    p_exclude_link = strdup(p_exclude_link);
}

char * get_exclude_list(int type)
{  
  switch(type) {
    case EXCLUDE_WRAP:
      return(p_exclude_wrap ? strdup(p_exclude_wrap) : NULL);
    case EXCLUDE_LINK:
      return(p_exclude_link ? strdup(p_exclude_link) : NULL);
    default:
      return(NULL);      
  }  
}

int is_excluded_list(char *p_name, EXCLUDE_TYPE type)
{
  char *p_list = get_exclude_list(type);
  int ret = FALSE;

  if(p_list) {
    char *p_tmp = p_list;
    char *p_file = tail_get(p_name, '/');
    char *p_act;
    while((p_act = strsep(&p_tmp, ":"))) {
      if(!fnmatch(p_act,p_file,0)) {
        ret = TRUE;
        break;
      }
    }
    free(p_list);
  }

  return(ret);
}

int print_excluded_list(EXCLUDE_TYPE type)
{
  char *p_list = get_exclude_list(type);
  if(p_list) {
    printf("%s:\n",exclude_names[type]);
    char *p_act;
    while((p_act = strsep(&p_list, ":"))) {
      printf("%s\n",p_act);
    }
    free(p_list);
  }
  return(FALSE);
}

void warning(const char *format, ...)
{
  if(cfg.verbose) {
    va_list args;
    
    fprintf(stderr, "WARNING: ");
    
    va_start(args, format);
    vfprintf(stderr, format, args);
    va_end(args);	
    
    fprintf(stderr, "\n");  
  }
}

void info(const char *format, ...)
{
  if(cfg.verbose) {
    va_list args;
    
    fprintf(stderr, "INFO: ");
    
    va_start(args, format);
    vfprintf(stderr, format, args);
    va_end(args);	
    
    fprintf(stderr, "\n");  
  }
}

void error(const char *format, ...)
{
  if(!cfg.quiet) {
    va_list args;
    
    fprintf(stderr, "ERROR: ");
    
    va_start(args, format);  
    vfprintf(stderr, format, args);
    va_end(args);
    
    fprintf(stderr, "\n");
  }  
	exit(1);
}

/*
 * Returns TRUE if SUID is active
 * e.g. process runs with root UID but it was lauched by user
 */
int is_setuid(void)
{
  return(geteuid() == 0 && getuid() != 0);
}

/*
 * Drop root UID
 */
int drop_setuid(void)
{  
  if(is_setuid()) {
    if(setgid(getgid()) < 0)
      return(FALSE);
    if(setuid(getuid()) < 0)
      return(FALSE);
  }  
  return(TRUE);
}

/*
 * Drop root UID, leave it only for filesystem
 */
int drop_root(void)
{
  if(is_setuid()) {
    return(drop_setuid());
  }
  
  return(TRUE);
}

enum
{
	EXIT_VIEWER_NOT_FOUND = -2,
	EXIT_VIEWER_ERROR = -1,
	EXIT_VIEWER_OK = 0,
};

char * get_prefix(char *p_prefix, int max_len, int s_bits, int t_bits)
{  
  snprintf(p_prefix,max_len,"nswrapper_%d_%d",s_bits,t_bits);
  return(p_prefix);
}

int check_plugin_viewer_bits(int bits)
{
  char *p_viewer = get_viewer_bits(bits);
  if(access(p_viewer, X_OK) != 0) {
    info("%s isn't accessible.",p_viewer);
    return(FALSE);
  }
  else {
    info("%s looks ok.",p_viewer);
    return(TRUE);
  }
}

/*
 * Check the wrapper and run a test for given plugin (if plugin_file isn't NULL)
 */
int check_plugin_viewer(const char *plugin_file, const char *p_arch)
{       
  char *p_viewer = get_viewer(p_arch);
  
  int pid = fork ();
  if (pid < 0) {
    return(FALSE);
  }
  else if (pid == 0) {
    if(!drop_setuid())
      exit(EXIT_VIEWER_ERROR);

    if(access(p_viewer, X_OK) != 0) {
      warning("%s isn't accessible!",p_viewer);
      exit(EXIT_VIEWER_ERROR);
    }
    
    execlp(p_viewer, p_viewer, "--test", "--plugin", plugin_file, NULL);
    warning("I can't find any viewer for plugin %s!",plugin_file);
    exit(EXIT_VIEWER_NOT_FOUND);
  }
  else {
    int status;    
    while (waitpid(pid, &status, 0) != pid)
      ;
    if(WIFEXITED(status)) {
      return(status == EXIT_VIEWER_OK);
    }
    warning("EXIT_VIEWER_ERROR for viewer %s and plug-in %s!",p_viewer,plugin_file);
    return(FALSE);
  }
}

int check_plugin_wrapper_bits(int bits)
{       
  char *p_wrapper = get_wrapper_bits(bits);
  if(access(p_wrapper, X_OK ) == 0) {
    return(TRUE);
  }
  else {
    warning("Wrapper %s is not accessible!",p_wrapper);
    return(FALSE);
  }
}

int check_plugin_wrapper(const char *p_arch)
{       
  return(check_plugin_wrapper_bits(get_bits(p_arch)));
}

int check_plugin(WRAP_PLUGIN *p_plugin, NPW_PLUGININFO * out_plugin_info)
{
	return(is_plugin(p_plugin->file.name, out_plugin_info) && 
         check_plugin_wrapper(out_plugin_info->target_arch) &&
         check_plugin_viewer(p_plugin->file.name, out_plugin_info->target_arch));
}

int get_wrap_config(WRAP_PLUGIN *p_plugin)
{
  int wrap = FALSE;

  if(p_plugin->t_bits == BITS_32 && p_plugin->s_bits == p_plugin->t_bits) {
    wrap = cfg.wrapper_32b;
  }
  if(p_plugin->t_bits == BITS_64 && p_plugin->s_bits == p_plugin->t_bits) {
    wrap = cfg.wrapper_64b;
  }
  if(p_plugin->t_bits == BITS_64 && p_plugin->s_bits == BITS_32) {
    wrap = cfg.wrapper_64b && cfg.wrapper_32b;
  }   
  
  return(wrap);
}

int get_link_config(WRAP_PLUGIN *p_plugin)
{
  int link = FALSE;

  if(p_plugin->t_bits == BITS_32 && p_plugin->s_bits == p_plugin->t_bits) {
    link = TRUE;
  }
  if(p_plugin->t_bits == BITS_64 && p_plugin->s_bits == p_plugin->t_bits) {
    link = TRUE;
  }
  
  return(link);
}

/**********************************************************************************
   Directory processing callback - plugins removing
  *********************************************************************************
*/
int remove_plugin(WRAP_PLUGIN *p_plugin, NPW_PLUGININFO * plugin_info)
{	
	info("Removing plugin %s...", p_plugin->file.name_wrapped);
	return(unlink(p_plugin->file.name_wrapped) == 0);
}

int remove_plugin_all(WRAP_PLUGIN *p_plugin, NPW_PLUGININFO * plugin_info)
{	
	info("Removing installed plugin %s...", p_plugin->file.name);
	unlink(p_plugin->file.name_wrapped);
	unlink(p_plugin->file.name_linked);
  return(TRUE);
}

/**********************************************************************************
   Directory processing callbacks - plugins installation
  *********************************************************************************
*/
int wrap_plugin_2(WRAP_PLUGIN *p_plugin, NPW_PLUGININFO * plugin_info)
{
  char *p_wrapper = get_wrapper_bits(p_plugin->t_bits);
  if(access(p_wrapper, X_OK) != 0) {
    warning("%s isn't accessible!",p_wrapper);
    return(FALSE);
  }
  
  int pid = fork ();
  if (pid < 0) {
    warning("Fork error, returned code %d",pid);
    return(FALSE);
  }
  else if (pid == 0) {
    char prefix[MAX_NAME];
    get_prefix(prefix, MAX_NAME, p_plugin->s_bits, p_plugin->t_bits);
    
    {
      int i;
      char *params[] = 
      {
        p_wrapper,
        "-n",
        "-p", 
        prefix,
        "-d", 
        p_plugin->dir.name_wrapped,
        "-i", 
        p_plugin->file.name,
        NULL
      };
    
      info("Running: %s", p_wrapper);
      for(i = 0; i < sizeof(params)/sizeof(params[0]); i++) {
        if(params[i]) {
          info("         %s",params[i]);
        }
      }
    
      execvp(p_wrapper, params);
    }
    
    warning("Wrapping error!");
    exit(1);
  }
  else {
    int status;
    while (waitpid(pid, &status, 0) != pid)
      ;
    return(WIFEXITED(status));
  }
}

// Wrap given plugin (if posibile)
int wrap_plugin(WRAP_PLUGIN *p_plugin, NPW_PLUGININFO * plugin_info)
{
  info("Wrapping %s to %s...",p_plugin->file.name, p_plugin->file.name_wrapped);
  if(check_plugin(p_plugin, plugin_info)) {
    return(wrap_plugin_2(p_plugin, plugin_info));
  }
  else {     
	  return(FALSE);
  }
}

// Only link given plugin
int link_plugin(WRAP_PLUGIN *p_plugin, NPW_PLUGININFO * plugin_info)
{  
  info("Linking %s to %s...",p_plugin->file.name, p_plugin->file.name_linked);
  return(symlink(p_plugin->file.name, p_plugin->file.name_linked) == 0);
}

// Wrap or link selected plug-in
int install_plugin(WRAP_PLUGIN *p_plugin, NPW_PLUGININFO * plugin_info)
{
  int wrap = get_wrap_config(p_plugin) && !is_excluded_list(p_plugin->file.name, EXCLUDE_WRAP);
  int link = get_link_config(p_plugin) && !is_excluded_list(p_plugin->file.name, EXCLUDE_LINK);
  
  if(wrap)
    wrap = wrap_plugin(p_plugin, plugin_info);
  
  if(!wrap && link)
    return(link_plugin(p_plugin, plugin_info));
  
  return(TRUE);
}

/**********************************************************************************
   Directory processing callbacks - plugins checking
  *********************************************************************************
*/
int check_installed_plugin_wrap(WRAP_PLUGIN *p_plugin, NPW_PLUGININFO * plugin_info)
{
  info("Check wrap %s to %s...",p_plugin->file.name, p_plugin->file.name_wrapped);
  
  if(access(p_plugin->file.name_wrapped, X_OK | R_OK) != 0) {
    info("FAIL: %s", p_plugin->file.name_wrapped);
    info("        It isn't installed.");
    return(FALSE);
  }
  
  if(!is_wrapper_plugin(p_plugin->file.name_wrapped,plugin_info)) {
    info("FAIL: %s", p_plugin->file.name_wrapped);
    info("        It doesn't look like a wrapped plugin.");
    return(FALSE);
  } else {
    info("OK:   %s",p_plugin->file.name_wrapped);
    info("        It's a wrapped plugin.");
    return(TRUE);
  }
}

// Only link given plugin
int check_installed_plugin_link(WRAP_PLUGIN *p_plugin, NPW_PLUGININFO * plugin_info)
{ 
  ssize_t len;
  char link_to[MAX_NAME];
  
  info("Check link %s to %s...",p_plugin->file.name, p_plugin->file.name_linked);
  
  len = readlink(p_plugin->file.name_linked, link_to, MAX_NAME-2);
  if(len < 0) {
    info("FAIL: %s",p_plugin->file.name_linked);
    info("        Readlink failed.");
    return(FALSE);
  }
  else {
    link_to[len] = '\0';
    if(strcmp(p_plugin->file.name,link_to)) {
      info("FAIL: %s",p_plugin->file.name_linked);
      info("        It is not a link to %s.",p_plugin->file.name);
      return(FALSE);
    }
    else {
      info("OK:   %s",p_plugin->file.name_linked);
      return(TRUE);
    }
  }
}

// Wrap or link selected plug-in
int check_installed_plugin(WRAP_PLUGIN *p_plugin, NPW_PLUGININFO * plugin_info)
{
  int wrap = get_wrap_config(p_plugin) && !is_excluded_list(p_plugin->file.name, EXCLUDE_WRAP); // can be wrapped
  int link = get_link_config(p_plugin) && !is_excluded_list(p_plugin->file.name, EXCLUDE_LINK); // can be linked
  int plug = is_plugin(p_plugin->file.name,plugin_info); // is it a correct plug-in?
  
  int wrap_current = check_installed_plugin_wrap(p_plugin, plugin_info);
  int link_current = check_installed_plugin_link(p_plugin, plugin_info);
  
  int wrap_ok = (wrap && wrap_current) || (!wrap && !wrap_current);
  int link_ok = (link && link_current) || (!link && !link_current);
  
  int ret = FALSE;
  
  // It's correct plugin
  // should be wrapped
  if(plug) {    
    if(wrap_ok && !link_ok) { // is wrapped?
      return(wrap);
    } else {
      ret = (wrap_ok && link_ok);
    }
  }
  // It is not correct plugin
  // should be linked
  else {
    if(wrap_current) {
      ret = FALSE;
    } else {
      ret = link_ok;
    }
  }
  
  if(!ret) {
    info("FAIL: %s",p_plugin->file.name);
  }
  
  info("        is correct plugin %d",plug);
  info("        can be wrapped %d, is wrapped %d",wrap,wrap_current);
  info("        can be linked %d, is linked  %d",link,link_current);    
    
  if(!ret && cfg.force) {
    info("      Trying to reinstall %s",p_plugin->file.name);
    remove_plugin_all(p_plugin, plugin_info);
    ret = install_plugin(p_plugin,plugin_info);
  }
  
  return(ret);
}

/**********************************************************************************
   Directory processing callback - Check plugin installation / symlink installation
  *********************************************************************************
*/
int check_plugin_installation(WRAP_PLUGIN *p_plugin, NPW_PLUGININFO * plugin_info)
{	
  static char link_to[MAX_NAME];
  ssize_t len;
  char *p_path = NULL;  
  
	info("Checking plugin installation %s...", p_plugin->file.name_wrapped);
  
  if(is_wrapper_plugin(p_plugin->file.name_wrapped,plugin_info)) {
    p_path = plugin_info->path;
  } else {    
    if((len = readlink(p_plugin->file.name_wrapped, link_to, MAX_NAME-2)) > 0) {
      link_to[len] = '\0';
      p_path = link_to;
    }
  }
  
  if(p_path && access(p_path, R_OK) != 0) {
    info("FAIL: %s", p_path);
    info("        Original plugin isn't installed.");
    return(unlink(p_plugin->file.name_wrapped) == 0);
  }
    
  return(TRUE);
}

/**********************************************************************************
   Directory processing callback - print plugins info
  *********************************************************************************
*/
int list_plugin(WRAP_PLUGIN *p_plugin, NPW_PLUGININFO * plugin_info)
{
	if(is_wrapper_plugin(p_plugin->file.name_wrapped, plugin_info)) {
    printf ("%s\n", p_plugin->file.name_wrapped);
    printf ("  Original plugin: %s\n", plugin_info->path);
    char *str = strtok (plugin_info->ident, ":");
    if (str && strcmp (str, "NPW") == 0){
      str = strtok (NULL, ":");
      if(str) {
        printf ("  Wrapper version string: %s", str);
        str = strtok (NULL, ":");
        if (str)
          printf (" (%s)", str);
        printf ("\n");
      }
    }
  }
  else {
    printf("File/Link %s\n",p_plugin->file.name_wrapped);
  }

	return(TRUE);
}

/**********************************************************************************
   Directory processing
  *********************************************************************************
*/
typedef int (*PROCESS_PLUGIN_CB) (WRAP_PLUGIN *p_plugin, NPW_PLUGININFO * p_plugin_info);

int process_plugin_dir(int s_bits, int t_bits,
                       const char *p_plugin_dir,
                       const char *p_plugin_dir_wrapped,
                       SCANNED_DIR scanned_dir,
                       PROCESS_PLUGIN_CB process)
{
  NPW_PLUGININFO  plugin_info;
  WRAP_PLUGIN     plugin;
	struct dirent  *ent;
  DIR            *dir;
  char            prefix[MAX_NAME];
  int             ret = TRUE;
  
  if(scanned_dir == PLUGIN_DIR) {
	  info("Processing %s -> %s", p_plugin_dir, p_plugin_dir_wrapped);      
	  dir = opendir(p_plugin_dir);
    if(dir == NULL) {
      warning("I can't open dir %s", p_plugin_dir_wrapped);
      return(TRUE);
    }
  }
  else {
    info("Processing %s <- %s", p_plugin_dir, p_plugin_dir_wrapped);
    dir = opendir(p_plugin_dir_wrapped);
    if(dir == NULL) {
      warning("I can't open dir %s", p_plugin_dir_wrapped);
      return(TRUE);
    }
  }

  strncpy(plugin.dir.name,p_plugin_dir,MAX_NAME);
  strncpy(plugin.dir.name_wrapped,p_plugin_dir_wrapped,MAX_NAME);
  strncpy(plugin.dir.name_linked,p_plugin_dir_wrapped,MAX_NAME);
  
  plugin.dir.name[MAX_NAME-1] = '\0';
  plugin.dir.name_wrapped[MAX_NAME-1] = '\0';
  plugin.dir.name_linked[MAX_NAME-1] = '\0';
  
  plugin.s_bits = s_bits;
  plugin.t_bits = t_bits;
  
  if(scanned_dir == PLUGIN_DIR)
    get_prefix(prefix,MAX_NAME,s_bits,t_bits);
	
	while ((ent = readdir(dir)) != NULL) {
    if(!is_excluded_default(ent->d_name)) {
      
      snprintf(plugin.file.name,MAX_NAME,"%s/%s",p_plugin_dir,ent->d_name);
      snprintf(plugin.file.name_linked,MAX_NAME,"%s/%s",p_plugin_dir_wrapped,ent->d_name);
      
      if(scanned_dir == PLUGIN_DIR)
        snprintf(plugin.file.name_wrapped,MAX_NAME,"%s/%s.%s",p_plugin_dir_wrapped,prefix,ent->d_name);
      else
        strncpy(plugin.file.name_wrapped,plugin.file.name_linked,MAX_NAME);
      
      if(!process(&plugin, &plugin_info)) {
        warning("Failed.");
        ret = FALSE;
      }
    }
	}
	
	closedir(dir);
	return(ret);
}

int process(PROCESS_PLUGIN_CB process, SCANNED_DIR scanned_dir)
{
  int ret = TRUE, ret2;
      
  if(cfg.config_32b) {
    WRAP_NAME *p_dir = get_plugin_dir_bits(BITS_32);
    ret2 = process_plugin_dir(BITS_32, BITS_32, p_dir->name, p_dir->name_wrapped, scanned_dir, process);
    if(!ret2)
      ret = FALSE;
  }
  if(cfg.config_64b) {
    WRAP_NAME *p_dir = get_plugin_dir_bits(BITS_64);
    ret2 = process_plugin_dir(BITS_64, BITS_64, p_dir->name, p_dir->name_wrapped, scanned_dir, process);
    if(!ret2)
      ret = FALSE;
  }    
  if(cfg.config_32b && cfg.config_64b) {
    WRAP_NAME *p_dir32 = get_plugin_dir_bits(BITS_32);
    WRAP_NAME *p_dir64 = get_plugin_dir_bits(BITS_64);
    ret2 = process_plugin_dir(BITS_32, BITS_64, p_dir32->name, p_dir64->name_wrapped, scanned_dir, process);
    if(!ret2)
      ret = FALSE;
  }  
  
  return(ret);
}

void configure(void)
{
  cfg.verbose = FALSE;
  cfg.quiet = FALSE;

  cfg.config_32b = TRUE;      // Configure plugins for 32bit browsers
  cfg.config_64b = TRUE;      // Configure plugins for 64bit browsers

  cfg.force = FALSE;          // Re-install all plug-ins

  cfg.wrapper_32b = FALSE;    // Wraper for for 32bit browsers is availabile
  cfg.wrapper_64b = FALSE;    // Wraper for for 64bit browsers is availabile 
  
  cfg.action = A_CHECK;

  create_exclude_lists();  

  umask(S_IWGRP|S_IWOTH); 
}

/**********************************************************************************
   Command-line option handlers
  *********************************************************************************/
void handle_install(void)
{  
  cfg.action = A_INSTALL;
}

void handle_remove(void)
{
  cfg.action = A_REMOVE;
}

void handle_check(void)
{
  cfg.action = A_CHECK;
}

void handle_list(void)
{
  cfg.action = A_LIST;
}

void handle_verbose(void)
{
	cfg.verbose = TRUE;
  cfg.quiet = FALSE;
}

void handle_quiet(void)
{
	cfg.quiet = TRUE;
  cfg.verbose = FALSE;
}

void handle_32b(void)
{
  cfg.config_32b = TRUE;
	cfg.config_64b = FALSE;
}

void handle_64b(void)
{
	cfg.config_64b = TRUE;
  cfg.config_32b = FALSE;
}

void handle_force(void)
{
	cfg.force = TRUE;
}

void handle_help(void)
{
	cfg.action = A_HELP;
}

int print_usage (void)
{
  printf("\
plugin-config (C) Red Hat, Inc. 2007. It's based on                       \n\
nspluginwrapper (C) 2005-2007 Gwenole Beauchesne                          \n\
                                                                          \n\
This program is free software; you can redistribute it and/or modify      \n\
it under the terms of the GNU General Public License as published by      \n\
the Free Software Foundation; either version 2 of the License, or         \n\
(at your option) any later version.                                       \n\
                                                                          \n\
This program links or wrap mozilla plugins from                           \n\
%s or %s \n\
to \n\
%s or %s \n\
                                                                          \n\
Usage:                                                                    \n\
                                                                          \n\
  plugin-setup  [options]                                                 \n\
                                                                          \n\
Options:                                                                  \n\
                                                                          \n\
  -i  --install Install/wrap all availabile plug-ins                      \n\
  -r  --remove  Remove all installed plug-ins                             \n\
                                                                          \n\
  -l  --list    List all installed plug-ins                               \n\
  -c  --check   Check all plug-ins                                        \n\
                                                                          \n\
  -f  --force   Used with '-i' removes all plugins before the installation\n\
                Used with '-c' tries to reinstall all broken plugins      \n\
                                                                          \n\
  -3  --32bit   Process only plugins for 32-bit browsers                  \n\
  -6  --64bit   Process only plugins for 64-bit browsers                  \n\
                                                                          \n\
  -v  --verbose Verbose mode                                              \n\
  -h  --help    This help                                                 \n\
                                                                          \n\
  If it is run without any command line argument, --check is used.        \n\\n",
  NPW_PLUGIN_DIR_32,NPW_PLUGIN_DIR_64,
  NPW_PLUGIN_DIR_32_WRAPPED,NPW_PLUGIN_DIR_64_WRAPPED );
  
  return(TRUE);                                                                          
}


typedef struct option {
  
  const char  *short_option;
  const char  *long_option;
  void       (*process_callback)(void);
  
} OPTION;

OPTION options[] =
{
  { "-i", "--install",  handle_install },
  { "-r", "--remove",   handle_remove },
  
  { "-l", "--list",     handle_list },
  { "-c", "--check",    handle_check },
  
  { "-f", "--force",    handle_force },
  
  { "-3", "--32bit",    handle_32b },
  { "-6", "--64bit",    handle_64b },
  
  { "-v", "--verbose",  handle_verbose },
  { "-h", "--help",     handle_help }
};

#define OPTION_SIZE   (sizeof(options)/sizeof(options[0]))

int main (int argc, char *argv[])
{
	int i, j, ret = TRUE;
  int set_uid = is_setuid();

  configure();
  
  for (i = 1; i < argc; i++)
  {
    for (j = 0; j < OPTION_SIZE; j++) {
      if(!strcmp(options[j].short_option,argv[i]) || !strcmp(options[j].long_option,argv[i])) {
         options[j].process_callback();
         break;
      }
    }
    if(i == OPTION_SIZE) {
      fprintf(stderr, "Wrong option - '%s'",argv[i]);
      print_usage();
      return(ret);
    }      
  }

  if(cfg.action == A_HELP) {    
    print_usage();
    return(0);
  }

  if(cfg.action == A_REMOVE && set_uid) {
    error("Only root can remove globally installed plugins");
  }
  
  cfg.wrapper_32b = check_plugin_viewer_bits(BITS_32) && check_plugin_wrapper_bits(BITS_32);
  cfg.wrapper_64b = check_plugin_viewer_bits(BITS_64) && check_plugin_wrapper_bits(BITS_64);
  
  switch(cfg.action) {
    case A_INSTALL:
    {
      if(cfg.force) {
        process(remove_plugin, PLUGIN_DIR_WRAPPED);
      }
      ret = process(install_plugin, PLUGIN_DIR);
      process(check_plugin_installation, PLUGIN_DIR_WRAPPED);
      break;
    }
    case A_REMOVE:
      ret = process(remove_plugin, PLUGIN_DIR_WRAPPED);
      break;
    case A_LIST:
      print_excluded_list(EXCLUDE_WRAP);
      print_excluded_list(EXCLUDE_LINK);
      ret = process(list_plugin, PLUGIN_DIR_WRAPPED);
      break;
    case A_CHECK:
      ret = process(check_installed_plugin, PLUGIN_DIR);
      if(!ret && cfg.force) {
        process(remove_plugin, PLUGIN_DIR_WRAPPED);
        ret = process(install_plugin, PLUGIN_DIR);
      }
      if(cfg.force) {
        process(check_plugin_installation, PLUGIN_DIR_WRAPPED);
      }
      break;
    default:
      break;
  }  
  
	return(!ret);
}
