/*
 *  plugin-detection.c - A plugin configuration tool
 *
 *  by Martin Stransky <stransky@redhat.com>
 *
 *  This utility is based on:
 *
 *  npw-config.c - nspluginwrapper configuration tool 
 *  nspluginwrapper (C) 2005-2007 Gwenole Beauchesne
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
	This utility will install plugins for mozilla based browsers or
  wrap them with the nspluginwrapper utility (if availabile)
*/

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <stdarg.h>
#include <limits.h>
#include <errno.h>

#include <dlfcn.h>
#include <unistd.h>
#include <fcntl.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <pwd.h>
#include <dirent.h>

#include <asm/types.h>

#include "plugin-detection.h"


/* ELF decoder derived from QEMU code */
#undef bswap_16
#define bswap_16(x) \
	((((x) >> 8) & 0xff) | (((x) & 0xff) << 8))

#undef bswap_32
#define bswap_32(x) \
     ((((x) & 0xff000000) >> 24) | (((x) & 0x00ff0000) >>  8) | \
      (((x) & 0x0000ff00) <<  8) | (((x) & 0x000000ff) << 24))

#undef bswap_64
#define bswap_64(x) \
     (              \
      (((x) & 0xff00000000000000) >> 56)  | (((x) & 0x00ff000000000000) >> 40) | \
      (((x) & 0x00000000000000ff) << 56)  | (((x) & 0x000000000000ff00) << 40) | \
      (((x) & 0x0000ff0000000000) >> 24)  | (((x) & 0x000000ff00000000) >>  8) | \
      (((x) & 0x0000000000ff0000) << 24)  | (((x) & 0x00000000ff000000) <<  8) \
     )

/* 32-bit ELF base types. */
typedef uint32_t Elf32_Addr;
typedef uint16_t Elf32_Half;
typedef uint32_t Elf32_Off;
typedef int32_t Elf32_Sword;
typedef uint32_t Elf32_Word;

/* 64-bit ELF base types. */
typedef uint64_t Elf64_Addr;
typedef uint16_t Elf64_Half;
typedef int16_t Elf64_SHalf;
typedef uint64_t Elf64_Off;
typedef int32_t Elf64_Sword;
typedef uint32_t Elf64_Word;
typedef uint64_t Elf64_Xword;
typedef int64_t Elf64_Sxword;

/* The ELF file header.  This appears at the start of every ELF file.  */
#define EI_NIDENT (16)

typedef struct
{
	unsigned char e_ident[EI_NIDENT];	/* Magic number and other info */
	Elf32_Half e_type;	/* Object file type */
	Elf32_Half e_machine;	/* Architecture */
	Elf32_Word e_version;	/* Object file version */
	Elf32_Addr e_entry;	/* Entry point virtual address */
	Elf32_Off e_phoff;	/* Program header table file offset */
	Elf32_Off e_shoff;	/* Section header table file offset */
	Elf32_Word e_flags;	/* Processor-specific flags */
	Elf32_Half e_ehsize;	/* ELF header size in bytes */
	Elf32_Half e_phentsize;	/* Program header table entry size */
	Elf32_Half e_phnum;	/* Program header table entry count */
	Elf32_Half e_shentsize;	/* Section header table entry size */
	Elf32_Half e_shnum;	/* Section header table entry count */
	Elf32_Half e_shstrndx;	/* Section header string table index */
} Elf32_Ehdr;

typedef struct elf64_hdr
{
	unsigned char e_ident[EI_NIDENT];	/* ELF "magic number" */
	Elf64_Half e_type;
	Elf64_Half e_machine;
	Elf64_Word e_version;
	Elf64_Addr e_entry;	/* Entry point virtual address */
	Elf64_Off e_phoff;	/* Program header table file offset */
	Elf64_Off e_shoff;	/* Section header table file offset */
	Elf64_Word e_flags;
	Elf64_Half e_ehsize;
	Elf64_Half e_phentsize;
	Elf64_Half e_phnum;
	Elf64_Half e_shentsize;
	Elf64_Half e_shnum;
	Elf64_Half e_shstrndx;
} Elf64_Ehdr;

/* Base structure - used to distinguish between 32/64 bit version */
typedef struct elf_hdr_base
{

	unsigned char e_ident[EI_NIDENT];	/* Magic number and other info */

} Elf_hdr_base;

#define EI_MAG0		0	/* File identification byte 0 index */
#define ELFMAG0		0x7f	/* Magic number byte 0 */
#define EI_MAG1		1	/* File identification byte 1 index */
#define ELFMAG1		'E'	/* Magic number byte 1 */
#define EI_MAG2		2	/* File identification byte 2 index */
#define ELFMAG2		'L'	/* Magic number byte 2 */
#define EI_MAG3		3	/* File identification byte 3 index */
#define ELFMAG3		'F'	/* Magic number byte 3 */
#define EI_CLASS	4	/* File class byte index */
#define ELFCLASS32	1	/* 32-bit objects */
#define ELFCLASS64	2	/* 64-bit objects */
#define EI_DATA		5	/* Data encoding byte index */
#define ELFDATA2LSB	1	/* 2's complement, little endian */
#define ELFDATA2MSB	2	/* 2's complement, big endian */
#define EI_OSABI	7	/* OS ABI identification */
#define ELFOSABI_SYSV	0	/* UNIX System V ABI */
#define ELFOSABI_NETBSD	2	/* NetBSD.  */
#define ELFOSABI_LINUX	3	/* Linux.  */
#define ELFOSABI_SOLARIS 6	/* Sun Solaris.  */
#define ELFOSABI_FREEBSD 9	/* FreeBSD.  */
#define ET_DYN		3	/* Shared object file */
#define EM_386		3	/* Intel 80386 */
#define EM_SPARC	2	/* SUN SPARC */
#define EM_PPC		20	/* PowerPC */
#define EM_PPC64	21	/* PowerPC 64-bit */
#define EM_SPARCV9	43	/* SPARC v9 64-bit */
#define EM_IA_64	50	/* Intel Merced */
#define EM_X86_64	62	/* AMD x86-64 architecture */
#define EV_CURRENT	1	/* Current version */

/* Section header.  */
typedef struct
{
	Elf32_Word sh_name;	/* Section name (string tbl index) */
	Elf32_Word sh_type;	/* Section type */
	Elf32_Word sh_flags;	/* Section flags */
	Elf32_Addr sh_addr;	/* Section virtual addr at execution */
	Elf32_Off sh_offset;	/* Section file offset */
	Elf32_Word sh_size;	/* Section size in bytes */
	Elf32_Word sh_link;	/* Link to another section */
	Elf32_Word sh_info;	/* Additional section information */
	Elf32_Word sh_addralign;	/* Section alignment */
	Elf32_Word sh_entsize;	/* Entry size if section holds table */
} Elf32_Shdr;

typedef struct elf64_shdr
{
	Elf64_Word sh_name;	/* Section name, index in string tbl */
	Elf64_Word sh_type;	/* Type of section */
	Elf64_Xword sh_flags;	/* Miscellaneous section attributes */
	Elf64_Addr sh_addr;	/* Section virtual addr at execution */
	Elf64_Off sh_offset;	/* Section file offset */
	Elf64_Xword sh_size;	/* Size of section in bytes */
	Elf64_Word sh_link;	/* Index of another section */
	Elf64_Word sh_info;	/* Additional section information */
	Elf64_Xword sh_addralign;	/* Section alignment */
	Elf64_Xword sh_entsize;	/* Entry size if section holds table */
} Elf64_Shdr;


#define SHT_NOBITS	  8	/* Program space with no data (bss) */
#define SHT_DYNSYM	  11	/* Dynamic linker symbol table */

/* Symbol table entry.  */
typedef struct
{
	Elf32_Word st_name;	/* Symbol name (string tbl index) */
	Elf32_Addr st_value;	/* Symbol value */
	Elf32_Word st_size;	/* Symbol size */
	unsigned char st_info;	/* Symbol type and binding */
	unsigned char st_other;	/* Symbol visibility */
	Elf32_Half st_shndx;	/* Section index */
} Elf32_Sym;

typedef struct elf64_sym
{
	Elf64_Word st_name;	/* Symbol name, index in string tbl */
	unsigned char st_info;	/* Type and binding attributes */
	unsigned char st_other;	/* No defined meaning, 0 */
	Elf64_Half st_shndx;	/* Associated section index */
	Elf64_Addr st_value;	/* Value of the symbol */
	Elf64_Xword st_size;	/* Associated symbol size */
} Elf64_Sym;

#define ELF_ST_BIND(x)          ((x) >> 4)
#define ELF_ST_TYPE(x)          (((unsigned int) x) & 0xf)
#define ELF32_ST_BIND(x)        ELF_ST_BIND(x)
#define ELF32_ST_TYPE(x)        ELF_ST_TYPE(x)
#define ELF64_ST_BIND(x)        ELF_ST_BIND(x)
#define ELF64_ST_TYPE(x)        ELF_ST_TYPE(x)

#define STB_GLOBAL	1	/* Global symbol */
#define STT_OBJECT	1	/* Symbol is a data object */
#define STT_FUNC	2	/* Symbol is a code object */

/* We handle 32-bit and 64-bit ELF plugins */
#define ElfW32(x)		Elf32_ ## x
#define ELFW32(x)		ELF32_ ## x

#define ElfW64(x)		Elf64_ ## x
#define ELFW64(x)		ELF64_ ## x

void * load_data (int fd, long offset, unsigned int size)
{
	char *data = (char *) malloc (size);
	if (!data)
		return NULL;

	lseek (fd, offset, SEEK_SET);
	if (read (fd, data, size) != size)
	{
		free (data);
		return NULL;
	}

	return data;
}

static int is_little_endian (void)
{
	union
	{
		uint32_t i;
		uint8_t b[4];
	} x;
	x.i = 0x01020304;
	return x.b[0] == 0x04;
}

static void elf_swap_ehdr_32 (ElfW32 (Ehdr) * hdr)
{
	hdr->e_type = bswap_16 (hdr->e_type);
	hdr->e_machine = bswap_16 (hdr->e_machine);
	hdr->e_version = bswap_32 (hdr->e_version);
	hdr->e_entry = bswap_32 (hdr->e_entry);
	hdr->e_phoff = bswap_32 (hdr->e_phoff);
	hdr->e_shoff = bswap_32 (hdr->e_shoff);
	hdr->e_flags = bswap_32 (hdr->e_flags);
	hdr->e_ehsize = bswap_16 (hdr->e_ehsize);
	hdr->e_phentsize = bswap_16 (hdr->e_phentsize);
	hdr->e_phnum = bswap_16 (hdr->e_phnum);
	hdr->e_shentsize = bswap_16 (hdr->e_shentsize);
	hdr->e_shnum = bswap_16 (hdr->e_shnum);
	hdr->e_shstrndx = bswap_16 (hdr->e_shstrndx);
}

static void elf_swap_ehdr_64 (ElfW64 (Ehdr) * hdr)
{
	hdr->e_type = bswap_16 (hdr->e_type);
	hdr->e_machine = bswap_16 (hdr->e_machine);
	hdr->e_version = bswap_32 (hdr->e_version);
	hdr->e_entry = bswap_64 (hdr->e_entry);
	hdr->e_phoff = bswap_64 (hdr->e_phoff);
	hdr->e_shoff = bswap_64 (hdr->e_shoff);
	hdr->e_flags = bswap_32 (hdr->e_flags);
	hdr->e_ehsize = bswap_16 (hdr->e_ehsize);
	hdr->e_phentsize = bswap_16 (hdr->e_phentsize);
	hdr->e_phnum = bswap_16 (hdr->e_phnum);
	hdr->e_shentsize = bswap_16 (hdr->e_shentsize);
	hdr->e_shnum = bswap_16 (hdr->e_shnum);
	hdr->e_shstrndx = bswap_16 (hdr->e_shstrndx);
}

static void elf_swap_shdr_32 (ElfW32 (Shdr) * shdr)
{
	shdr->sh_name = bswap_32 (shdr->sh_name);
	shdr->sh_type = bswap_32 (shdr->sh_type);
	shdr->sh_flags = bswap_32 (shdr->sh_flags);
	shdr->sh_addr = bswap_32 (shdr->sh_addr);
	shdr->sh_offset = bswap_32 (shdr->sh_offset);
	shdr->sh_size = bswap_32 (shdr->sh_size);
	shdr->sh_link = bswap_32 (shdr->sh_link);
	shdr->sh_info = bswap_32 (shdr->sh_info);
	shdr->sh_addralign = bswap_32 (shdr->sh_addralign);
	shdr->sh_entsize = bswap_32 (shdr->sh_entsize);
}

static void elf_swap_shdr_64 (ElfW64 (Shdr) * shdr)
{
	shdr->sh_name = bswap_32 (shdr->sh_name);
	shdr->sh_type = bswap_32 (shdr->sh_type);
	shdr->sh_flags = bswap_64 (shdr->sh_flags);
	shdr->sh_addr = bswap_64 (shdr->sh_addr);
	shdr->sh_offset = bswap_64 (shdr->sh_offset);
	shdr->sh_size = bswap_64 (shdr->sh_size);
	shdr->sh_link = bswap_32 (shdr->sh_link);
	shdr->sh_info = bswap_32 (shdr->sh_info);
	shdr->sh_addralign = bswap_64 (shdr->sh_addralign);
	shdr->sh_entsize = bswap_64 (shdr->sh_entsize);
}

static void elf_swap_sym_32 (ElfW32 (Sym) * sym)
{
	sym->st_name = bswap_32 (sym->st_name);
	sym->st_value = bswap_32 (sym->st_value);
	sym->st_size = bswap_32 (sym->st_size);
	sym->st_shndx = bswap_16 (sym->st_shndx);
}

static void elf_swap_sym_64 (ElfW64 (Sym) * sym)
{
	sym->st_name = bswap_32 (sym->st_name);
	sym->st_value = bswap_64 (sym->st_value);
	sym->st_size = bswap_64 (sym->st_size);
	sym->st_shndx = bswap_16 (sym->st_shndx);
}

int is_plugin_fd_32(int fd, NPW_PLUGININFO * out_plugin_info)
{
	int i;
	int ret = FALSE;

	ElfW32 (Ehdr) ehdr;
	if (read (fd, &ehdr, sizeof (ehdr)) != sizeof (ehdr))
		return FALSE;

	int do_swap = (ehdr.e_ident[EI_DATA] == ELFDATA2LSB)
		&& !is_little_endian ();
	if (do_swap)
		elf_swap_ehdr_32 (&ehdr);

	if (ehdr.e_type != ET_DYN)
		return FALSE;
	if (ehdr.e_version != EV_CURRENT)
		return FALSE;

	if (out_plugin_info)
	{
		const char *target_arch = "";
		switch (ehdr.e_machine)
		{
		case EM_386:
			target_arch = "i386";
			break;
		case EM_SPARC:
			target_arch = "sparc";
			break;
		case EM_PPC:
			target_arch = "ppc";
			break;
		}
		strcpy (out_plugin_info->target_arch, target_arch);
		const char *target_os = "";
		switch (ehdr.e_ident[EI_OSABI])
		{
		case ELFOSABI_LINUX:
			target_os = "linux";
			break;
		case ELFOSABI_SOLARIS:
			target_os = "solaris";
			break;
		case ELFOSABI_FREEBSD:
			target_os = "freebsd";
			break;
		}
		strcpy (out_plugin_info->target_os, target_os);
	}

	ElfW32 (Shdr) * shdr =
		(ElfW32 (Shdr) *) load_data (fd, ehdr.e_shoff,
					     ehdr.e_shnum * sizeof (*shdr));
	if (do_swap)
	{
		for (i = 0; i < ehdr.e_shnum; i++)
			elf_swap_shdr_32 (&shdr[i]);
	}

	char **sdata = (char **) calloc (ehdr.e_shnum, sizeof (*sdata));
	for (i = 0; i < ehdr.e_shnum; i++)
	{
		ElfW32 (Shdr) * sec = &shdr[i];
		if (sec->sh_type != SHT_NOBITS)
			sdata[i] =
				(char *) load_data (fd, sec->sh_offset,
						    sec->sh_size);
	}

	ElfW32 (Shdr) * symtab_sec = NULL;
	for (i = 0; i < ehdr.e_shnum; i++)
	{
		ElfW32 (Shdr) * sec = &shdr[i];
		if (sec->sh_type == SHT_DYNSYM
		    && strcmp (sdata[ehdr.e_shstrndx] + sec->sh_name,
			       ".dynsym") == 0)
		{
			symtab_sec = sec;
			break;
		}
	}
	if (symtab_sec == NULL)
		goto done;
	ElfW32 (Sym) * symtab = (ElfW32 (Sym) *) sdata[symtab_sec - shdr];
	char *strtab = sdata[symtab_sec->sh_link];

	int nb_syms = symtab_sec->sh_size / sizeof (*symtab);
	if (do_swap)
	{
		for (i = 0; i < nb_syms; i++)
			elf_swap_sym_32 (&symtab[i]);
	}

	int nb_np_syms;
	int is_wrapper_plugin = 0;
	for (i = 0, nb_np_syms = 0; i < nb_syms; i++)
	{
		ElfW32 (Sym) * sym = &symtab[i];
		const char *name = strtab + sym->st_name;
		if (ELFW32 (ST_BIND) (sym->st_info) != STB_GLOBAL)
			continue;
		if (ELFW32 (ST_TYPE) (sym->st_info) == STT_OBJECT
		    && strcmp (name, "NPW_Plugin") == 0)
			is_wrapper_plugin = 1;
		if (ELFW32 (ST_TYPE) (sym->st_info) != STT_FUNC)
			continue;
		if (!strcmp (name, "NP_GetMIMEDescription") ||
		    !strcmp (name, "NP_Initialize")
		    || !strcmp (name, "NP_Shutdown"))
			nb_np_syms++;
	}
	ret = (nb_np_syms == 3) && !is_wrapper_plugin;

      done:
	for (i = 0; i < ehdr.e_shnum; i++)
		free (sdata[i]);
	free (sdata);
	free (shdr);
	return ret;
}

int is_plugin_fd_64 (int fd, NPW_PLUGININFO * out_plugin_info)
{
	int i;
	int ret = FALSE;

	ElfW64 (Ehdr) ehdr;
	if (read (fd, &ehdr, sizeof (ehdr)) != sizeof (ehdr))
		return FALSE;

	int do_swap = (ehdr.e_ident[EI_DATA] == ELFDATA2LSB)
		&& !is_little_endian ();
	if (do_swap)
		elf_swap_ehdr_64 (&ehdr);

	if (ehdr.e_type != ET_DYN)
		return FALSE;
	if (ehdr.e_version != EV_CURRENT)
		return FALSE;

	if (out_plugin_info)
	{
		const char *target_arch = "";
		switch (ehdr.e_machine)
		{
		case EM_X86_64:
			target_arch = "x86_64";
			break;
		case EM_PPC64:
			target_arch = "ppc64";
			break;
		}
		strcpy (out_plugin_info->target_arch, target_arch);
		const char *target_os = "";
		switch (ehdr.e_ident[EI_OSABI])
		{
		case ELFOSABI_LINUX:
			target_os = "linux";
			break;
		case ELFOSABI_SOLARIS:
			target_os = "solaris";
			break;
		case ELFOSABI_FREEBSD:
			target_os = "freebsd";
			break;
		}
		strcpy (out_plugin_info->target_os, target_os);
	}

	ElfW64 (Shdr) * shdr =
		(ElfW64 (Shdr) *) load_data (fd, ehdr.e_shoff,
					     ehdr.e_shnum * sizeof (*shdr));
	if (do_swap)
	{
		for (i = 0; i < ehdr.e_shnum; i++)
			elf_swap_shdr_64 (&shdr[i]);
	}

	char **sdata = (char **) calloc (ehdr.e_shnum, sizeof (*sdata));
	for (i = 0; i < ehdr.e_shnum; i++)
	{
		ElfW64 (Shdr) * sec = &shdr[i];
		if (sec->sh_type != SHT_NOBITS)
			sdata[i] =
				(char *) load_data (fd, sec->sh_offset,
						    sec->sh_size);
	}

	ElfW64 (Shdr) * symtab_sec = NULL;
	for (i = 0; i < ehdr.e_shnum; i++)
	{
		ElfW64 (Shdr) * sec = &shdr[i];
		if (sec->sh_type == SHT_DYNSYM
		    && strcmp (sdata[ehdr.e_shstrndx] + sec->sh_name,
			       ".dynsym") == 0)
		{
			symtab_sec = sec;
			break;
		}
	}
	if (symtab_sec == NULL)
		goto done;
	ElfW64 (Sym) * symtab = (ElfW64 (Sym) *) sdata[symtab_sec - shdr];
	char *strtab = sdata[symtab_sec->sh_link];

	int nb_syms = symtab_sec->sh_size / sizeof (*symtab);
	if (do_swap)
	{
		for (i = 0; i < nb_syms; i++)
			elf_swap_sym_64 (&symtab[i]);
	}

	int nb_np_syms;
	int is_wrapper_plugin = 0;
	for (i = 0, nb_np_syms = 0; i < nb_syms; i++)
	{
		ElfW64 (Sym) * sym = &symtab[i];
		const char *name = strtab + sym->st_name;
		if (ELFW64 (ST_BIND) (sym->st_info) != STB_GLOBAL)
			continue;
		if (ELFW64 (ST_TYPE) (sym->st_info) == STT_OBJECT
		    && strcmp (name, "NPW_Plugin") == 0)
			is_wrapper_plugin = 1;
		if (ELFW64 (ST_TYPE) (sym->st_info) != STT_FUNC)
			continue;
		if (!strcmp (name, "NP_GetMIMEDescription") ||
		    !strcmp (name, "NP_Initialize")
		    || !strcmp (name, "NP_Shutdown"))
			nb_np_syms++;
	}
	ret = (nb_np_syms == 3) && !is_wrapper_plugin;

      done:
	for (i = 0; i < ehdr.e_shnum; i++)
		free (sdata[i]);
	free (sdata);
	free (shdr);
	return ret;
}

int is_plugin_fd (int fd, NPW_PLUGININFO * out_plugin_info)
{
	Elf_hdr_base ehdr;

	if (read (fd, &ehdr, sizeof (ehdr)) != sizeof (ehdr))
		return FALSE;

	if (ehdr.e_ident[EI_MAG0] != ELFMAG0
	    || ehdr.e_ident[EI_MAG1] != ELFMAG1
	    || ehdr.e_ident[EI_MAG2] != ELFMAG2
	    || ehdr.e_ident[EI_MAG3] != ELFMAG3)
		return FALSE;

	lseek (fd, 0, SEEK_SET);

	if (ehdr.e_ident[EI_CLASS] == ELFCLASS32)
		return (is_plugin_fd_32 (fd, out_plugin_info));
	else if (ehdr.e_ident[EI_CLASS] == ELFCLASS64)
		return (is_plugin_fd_64 (fd, out_plugin_info));

	return FALSE;
}

int is_plugin (const char *filename, NPW_PLUGININFO * out_plugin_info)
{
	int fd = open (filename, O_RDONLY);
	if (fd < 0)
		return FALSE;

	int ret = is_plugin_fd (fd, out_plugin_info);
	close (fd);
	return ret;
}

int is_wrapper_plugin_fd_32(int fd)
{
	int i;
	int ret = FALSE;

	ElfW32 (Ehdr) ehdr;
	if (read (fd, &ehdr, sizeof (ehdr)) != sizeof (ehdr))
		return FALSE;

	int do_swap = (ehdr.e_ident[EI_DATA] == ELFDATA2LSB)
		&& !is_little_endian ();
	if (do_swap)
		elf_swap_ehdr_32 (&ehdr);

	if (ehdr.e_type != ET_DYN)
		return FALSE;
	if (ehdr.e_version != EV_CURRENT)
		return FALSE;

	ElfW32 (Shdr) * shdr =
		(ElfW32 (Shdr) *) load_data (fd, ehdr.e_shoff,
					     ehdr.e_shnum * sizeof (*shdr));
	if (do_swap)
	{
		for (i = 0; i < ehdr.e_shnum; i++)
			elf_swap_shdr_32 (&shdr[i]);
	}

	char **sdata = (char **) calloc (ehdr.e_shnum, sizeof (*sdata));
	for (i = 0; i < ehdr.e_shnum; i++)
	{
		ElfW32 (Shdr) * sec = &shdr[i];
		if (sec->sh_type != SHT_NOBITS)
			sdata[i] =
				(char *) load_data (fd, sec->sh_offset,
						    sec->sh_size);
	}

	ElfW32 (Shdr) * symtab_sec = NULL;
	for (i = 0; i < ehdr.e_shnum; i++)
	{
		ElfW32 (Shdr) * sec = &shdr[i];
		if (sec->sh_type == SHT_DYNSYM
		    && strcmp (sdata[ehdr.e_shstrndx] + sec->sh_name,
			       ".dynsym") == 0)
		{
			symtab_sec = sec;
			break;
		}
	}
	if (symtab_sec == NULL)
		goto done;
	ElfW32 (Sym) * symtab = (ElfW32 (Sym) *) sdata[symtab_sec - shdr];
	char *strtab = sdata[symtab_sec->sh_link];

	int nb_syms = symtab_sec->sh_size / sizeof (*symtab);
	if (do_swap)
	{
		for (i = 0; i < nb_syms; i++)
			elf_swap_sym_32 (&symtab[i]);
	}
  
	int nb_np_syms;
	int is_wrapper_plugin = 0;
	for (i = 0, nb_np_syms = 0; i < nb_syms; i++)
	{
		ElfW32 (Sym) * sym = &symtab[i];
		const char *name = strtab + sym->st_name;
		if (ELFW32 (ST_BIND) (sym->st_info) != STB_GLOBAL)
			continue;
		if (ELFW32 (ST_TYPE) (sym->st_info) == STT_OBJECT && strcmp (name, "NPW_Plugin") == 0)
			is_wrapper_plugin = 1;
		if (ELFW32 (ST_TYPE) (sym->st_info) != STT_FUNC)
			continue;
		if (!strcmp(name, "NP_GetMIMEDescription") || !strcmp(name, "NP_Initialize") || !strcmp(name, "NP_Shutdown"))
			nb_np_syms++;
	}
	ret = (nb_np_syms == 3) && is_wrapper_plugin;
  
done:
	for (i = 0; i < ehdr.e_shnum; i++)
		free (sdata[i]);
	free (sdata);
	free (shdr);
	return ret;
}

int is_wrapper_plugin_fd_64 (int fd)
{
	int i;
	int ret = FALSE;

	ElfW64 (Ehdr) ehdr;
	if (read (fd, &ehdr, sizeof (ehdr)) != sizeof (ehdr))
		return FALSE;

	int do_swap = (ehdr.e_ident[EI_DATA] == ELFDATA2LSB)
		&& !is_little_endian ();
	if (do_swap)
		elf_swap_ehdr_64 (&ehdr);

	if (ehdr.e_type != ET_DYN)
		return FALSE;
	if (ehdr.e_version != EV_CURRENT)
		return FALSE;

	ElfW64 (Shdr) * shdr =
		(ElfW64 (Shdr) *) load_data (fd, ehdr.e_shoff,
					     ehdr.e_shnum * sizeof (*shdr));
	if (do_swap)
	{
		for (i = 0; i < ehdr.e_shnum; i++)
			elf_swap_shdr_64 (&shdr[i]);
	}

	char **sdata = (char **) calloc (ehdr.e_shnum, sizeof (*sdata));
	for (i = 0; i < ehdr.e_shnum; i++)
	{
		ElfW64 (Shdr) * sec = &shdr[i];
		if (sec->sh_type != SHT_NOBITS)
			sdata[i] =
				(char *) load_data (fd, sec->sh_offset,
						    sec->sh_size);
	}

	ElfW64 (Shdr) * symtab_sec = NULL;
	for (i = 0; i < ehdr.e_shnum; i++)
	{
		ElfW64 (Shdr) * sec = &shdr[i];
		if (sec->sh_type == SHT_DYNSYM
		    && strcmp (sdata[ehdr.e_shstrndx] + sec->sh_name,
			       ".dynsym") == 0)
		{
			symtab_sec = sec;
			break;
		}
	}
	if (symtab_sec == NULL)
		goto done;
	ElfW64 (Sym) * symtab = (ElfW64 (Sym) *) sdata[symtab_sec - shdr];
	char *strtab = sdata[symtab_sec->sh_link];

	int nb_syms = symtab_sec->sh_size / sizeof (*symtab);
	if (do_swap)
	{
		for (i = 0; i < nb_syms; i++)
			elf_swap_sym_64 (&symtab[i]);
	}

	int nb_np_syms;
	int is_wrapper_plugin = 0;
	for (i = 0, nb_np_syms = 0; i < nb_syms; i++)
	{
		ElfW64 (Sym) * sym = &symtab[i];
		const char *name = strtab + sym->st_name;
		if (ELFW64 (ST_BIND) (sym->st_info) != STB_GLOBAL)
			continue;
		if (ELFW64 (ST_TYPE) (sym->st_info) == STT_OBJECT && strcmp (name, "NPW_Plugin") == 0)
			is_wrapper_plugin = 1;
		if (ELFW64 (ST_TYPE) (sym->st_info) != STT_FUNC)
			continue;
		if (!strcmp (name, "NP_GetMIMEDescription") ||
		    !strcmp (name, "NP_Initialize")
		    || !strcmp (name, "NP_Shutdown"))
			nb_np_syms++;
	}
	ret = (nb_np_syms == 3) && is_wrapper_plugin;

  done:
	for (i = 0; i < ehdr.e_shnum; i++)
		free (sdata[i]);
	free (sdata);
	free (shdr);
	return ret;
}

int is_wrapper_plugin_fd (int fd)
{
	Elf_hdr_base ehdr;

	if (read (fd, &ehdr, sizeof (ehdr)) != sizeof (ehdr))
		return FALSE;

	if (ehdr.e_ident[EI_MAG0] != ELFMAG0
	    || ehdr.e_ident[EI_MAG1] != ELFMAG1
	    || ehdr.e_ident[EI_MAG2] != ELFMAG2
	    || ehdr.e_ident[EI_MAG3] != ELFMAG3)
		return FALSE;

	lseek (fd, 0, SEEK_SET);

	if(ehdr.e_ident[EI_CLASS] == ELFCLASS32)
		return(is_wrapper_plugin_fd_32(fd));
	else if(ehdr.e_ident[EI_CLASS] == ELFCLASS64)
		return(is_wrapper_plugin_fd_64(fd));

	return FALSE;
}

int get_wrapper_plugin_info(const char *plugin_path, NPW_PLUGININFO *out_plugin_info)
{
  void *handle = dlopen(plugin_path, RTLD_LAZY);
  if (handle == NULL)
        return FALSE;
  if (dlsym(handle, "NP_Initialize") == NULL)
        return FALSE;
  if (dlsym(handle, "NP_Shutdown") == NULL)
        return FALSE;
  if (dlsym(handle, "NP_GetMIMEDescription") == NULL)
        return FALSE;
  NPW_PLUGININFO *pi;
  if ((pi = (NPW_PLUGININFO *)dlsym(handle, "NPW_Plugin")) == NULL)
        return FALSE;
  if (out_plugin_info) {
        strncpy(out_plugin_info->ident, pi->ident, NPW_PLUGIN_IDENT_SIZE);
        strncpy(out_plugin_info->path, pi->path, PATH_MAX);
        out_plugin_info->mtime = pi->mtime;
        out_plugin_info->target_arch[0] = '\0';
        out_plugin_info->target_os[0] = '\0';
        if (strncmp(pi->ident, "NPW:0.9.9", 9) != 0) {
          strncpy(out_plugin_info->target_arch, pi->target_arch, NPW_PLUGININFO_TARGET_LENGTH-2);
          strncpy(out_plugin_info->target_os, pi->target_os, NPW_PLUGININFO_TARGET_LENGTH-2);
          out_plugin_info->target_arch[NPW_PLUGININFO_TARGET_LENGTH-1] = '\0';
          out_plugin_info->target_os[NPW_PLUGININFO_TARGET_LENGTH-1] = '\0';
        }
  }
  dlclose(handle);
  return TRUE;
}

int is_wrapper_plugin (const char *plugin_path, NPW_PLUGININFO * out_plugin_info)
{
	int fd = open (plugin_path, O_RDONLY);
	if (fd < 0)
		return FALSE;
	int ret = is_wrapper_plugin_fd(fd);	
  close (fd);
  
  if(out_plugin_info)
    get_wrapper_plugin_info(plugin_path, out_plugin_info);
  
	return ret;
}
